@@ -1,4 +1,4 @@
-#![doc(html_root_url = "https://docs.rs/singleton-cell/0.1.1")]
+#![doc(html_root_url = "https://docs.rs/singleton-cell/0.2.0")]
 #![no_std]
 
 use core::cell::UnsafeCell;
@@ -7,6 +7,13 @@ use core::mem;
 use singleton_trait::Exists;
 pub use singleton_trait::Singleton;
 
+pub mod internal {
+    /**
+     * Re-export paste macro so that we can refer to it in new_singleton
+     */
+    pub use paste;
+}
+
 /**
  * SCell, or SingletonCell provides an interface of a
  * Ghost Cell, where the Key is allowed to be any singleton,
@@ -98,7 +105,6 @@ impl<Key: ?Sized, T> SCell<Key, T> {
     }
 }
 
-
 impl<Key: ?Sized, T> SCell<Key, [T]> {
     /**
      * Returns a slice of cells from a cell of a slice
@@ -142,6 +148,7 @@ impl<Key: Singleton + ?Sized, T: ?Sized> SCell<Key, T> {
      * Safety to borrow mutably is guaranteed by the mutable borrow to the
      * singleton key for this cell.
      */
+    #[allow(clippy::mut_from_ref)]
     #[inline]
     pub fn borrow_mut<'a>(&'a self, _: impl Exists<&'a mut Key>) -> &'a mut T
     where
@@ -230,10 +237,88 @@ pub fn with_token<R>(f: impl for<'brand> FnOnce(Token<'brand>) -> R) -> R {
     f(token)
 }
 
+///
+/// Generates a new global singleton type with a fallible constructor.
+/// This is intended for cases where a single instance is created early
+/// in the program and passed throughout.
+///
+/// This is one way to create a Singleton which is 'static and Send/Sync
+///
+/// ```
+/// #[macro_use] extern crate singleton_cell;
+/// new_singleton!(pub Token);
+/// fn main() {
+///     let tok = Token::new().expect("Main called only once");
+///
+///     // possibly combine with Erased borrows
+///     use singleton_trait::Erased;
+///     let tok = Erased::new(tok);
+///
+/// }
+///
+/// ```
+///
+/// This method will return Some the first time it is called, and None
+/// any time thereafter. This does not establish any ordering between
+/// multiple threads if they access at the same time.
+///
+// Credit to Yandros for significant help with the idea / designing
+// the original macro using std::sync::Once instead of AtomicBool.
+//
+// We use AtomicBool here instead to remain no-std and simplify the
+// resulting code given that we don't need most features of Once.
+#[macro_export]
+macro_rules! new_singleton {
+    ( $pub:vis $Name:ident ) => {
+        $crate::internal::paste::paste! {
+            $pub use [< __ $Name _help__ >]::$Name;
+            #[allow(nonstandard_style)]
+            mod [< __ $Name _help__ >] {
+
+                pub
+                struct $Name(());
+
+                impl $Name {
+                    pub
+                    fn new () -> Option<Self> {
+                        let first_run = {
+                            static ONCE: ::core::sync::atomic::AtomicBool =
+                                ::core::sync::atomic::AtomicBool::new(false);
+                            &ONCE
+                        };
+                        if first_run
+                            // We don't need any ordering guarantees between accesses, only atomicity
+                            .compare_exchange(false, true,
+                                ::core::sync::atomic::Ordering::Relaxed,
+                                ::core::sync::atomic::Ordering::Relaxed)
+                            .is_ok()
+                        {
+                            // Safety:
+                            // * ONCE is a static, hence it is a single memory location which is false at
+                            // program start.
+                            // * The program start happens before all reads
+                            // * Only one CAS from false to true can succeed since it will never become
+                            // false again
+                            Some( $Name(()) )
+                        } else {
+                            None
+                        }
+                    }
+                }
+
+                // Safety: construction is guarded by the CAS on a static above
+                // So only one instance can be created per program
+                unsafe
+                impl ::singleton_trait::Singleton for $Name {}
+            }
+        }
+    };
+}
 
 #[cfg(test)]
 mod tests {
     extern crate alloc;
+    extern crate std;
     use super::*;
     use alloc::string::String;
 
@@ -265,7 +350,6 @@ mod tests {
 
     #[test]
     fn ghost_read_after_write() {
-
         let result = with_token(|mut tok| {
             let mut s = String::from("A");
             let s = SCell::from_mut(&mut s);
@@ -303,4 +387,21 @@ mod tests {
         assert_eq!("A!!C", s_mut);
     }
 
+    #[test]
+    fn global_singleton_unique() {
+        new_singleton!(pub A);
+        let _ = A::new();
+        assert!(A::new().is_none(), "Second call to A::new() should be None");
+    }
+
+    #[test]
+    fn global_singleton_distinct() {
+        new_singleton!(pub A);
+        new_singleton!(pub B);
+        let a: Option<A> = A::new();
+        let b: Option<B> = B::new();
+
+        assert!(a.is_some(), "A should be Some");
+        assert!(b.is_some(), "B should be Some");
+    }
 }
